 /*****************************************************************************
 *   Copyright (C) 2004 by Andrew de Quincey                                  *
 *   adq@lidskialf.net                                                        *
 *   Copyright (C) 2005 by Michael Schulze                                    *
 *   mike.s@genion.de                                                         *
 *                                                                            *
 *  The code contained in this file is free software; you can redistribute    *
 *  it and/or modify it under the terms of the GNU Lesser General Public      *
 *  License as published by the Free Software Foundation; either version      *
 *  2.1 of the License, or (at your option) any later version.                *
 *                                                                            *
 *  This file is distributed in the hope that it will be useful,              *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 *  Lesser General Public License for more details.                           *
 *                                                                            *
 *  You should have received a copy of the GNU Lesser General Public          *
 *  License along with this code; if not, write to the Free Software          *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *                                                                            *
 *  iTunes and iPod are trademarks of Apple                                   *
 *                                                                            *
 *  This product is not supported/written/published by Apple!                 *
 *****************************************************************************/

#include "ipoddevicedetails.h"
#include <qtextstream.h>

#include "utils.h"

#define MAX_DETAILS_STRING_LEN 255


IPodDeviceDetails::IPodDeviceDetails(const QString& filename) {
    this->filename = filename;
    dirty = false;
}

IPodDeviceDetails::~IPodDeviceDetails() {
}

bool IPodDeviceDetails::load() {
    QFile detailsFile( filename );
    if ( detailsFile.exists() && detailsFile.open( IO_ReadOnly ) ) {
        QStringList details = readDeviceInfos( detailsFile );
        switch ( details.count() ) {
        default:
        case 3:
            comment = details[ 2 ];
        case 2:
            owner = details[ 1 ];
        case 1:
            name = details[ 0 ];
        case 0:
            break;
        }
        detailsFile.close();
        dirty = false;
        return true;
    }

    // if we failed, mark it as dirty so we try and save a new one
    dirty = true;
    return false;
}

bool IPodDeviceDetails::save() {
    if ( !dirty ) {
        return true;
    }

    QFile detailsFile(filename);

    if (detailsFile.open(IO_WriteOnly)) {
        unsigned int length;
        QTextStream ostream(&detailsFile);
        ostream.setEncoding(QTextStream::RawUnicode);

        length = name.length();
        detailsFile.at(0x000);
        detailsFile.putch(length & 0xff);
        detailsFile.putch(length >> 8);
        ostream << name;

        length = owner.length();
        detailsFile.at(0x200);
        detailsFile.putch(length & 0xff);
        detailsFile.putch(length >> 8);
        ostream << owner;

        length = comment.length();
        detailsFile.at(0x400);
        detailsFile.putch(length & 0xff);
        detailsFile.putch(length >> 8);
        ostream << comment;

        // ensure file is the correct length
        detailsFile.at(0x5ff);
        detailsFile.putch(0);

        detailsFile.close();
        dirty = false;
        return true;
    }

    // failure!
    return false;
}

void IPodDeviceDetails::clear() {
    name = QString();
    owner = QString();
    comment = QString();
}

const QString& IPodDeviceDetails::getName() const {
    return name;
}

const QString& IPodDeviceDetails::getOwner() const {
    return owner;
}

const QString& IPodDeviceDetails::getComment() const {
    return comment;
}

void IPodDeviceDetails::setName(const QString& name) {
    this->name = name;
    this->name.truncate(MAX_DETAILS_STRING_LEN);
    dirty = true;
}

void IPodDeviceDetails::setOwner(const QString& owner) {
    this->owner = owner;
    this->owner.truncate(MAX_DETAILS_STRING_LEN);
    dirty = true;
}

void IPodDeviceDetails::setComment(const QString& comment) {
    this->comment = comment;
    this->comment.truncate(MAX_DETAILS_STRING_LEN);
    dirty = true;
}

QStringList IPodDeviceDetails::readDeviceInfos( QFile& file ) {
    QStringList result;

    QDataStream stream( &file );
    stream.setByteOrder( QDataStream::LittleEndian );

    for ( int i = 0; i < 3 && !stream.atEnd(); i++ ) {
        Q_UINT16 stringsize = 0;
        stream >> stringsize;
        unsigned int ucs2len = stringsize * 2;
        Q_UINT16 * buffer = new Q_UINT16 [ ucs2len + 1 ];

        for ( uint i = 0; i < ucs2len; ++i ) {
            stream >> buffer[ i ];
        }

        buffer[ ucs2len ] = 0;

        result.append( QString::fromUcs2( buffer ) );

        itunesdb::utils::seekRelative( stream, 198 - ucs2len );    // skip stuff

        delete [] buffer;
    }

    return result;
}

/*
QString IPodDeviceDetails::readDeviceInfoString(QFile& file, int position) {
    // get the string size
    file.at(position);
    unsigned int stringSize = file.getch() | (file.getch() << 8);
    if (stringSize > MAX_DETAILS_STRING_LEN) return "";

    // read in the raw data
    QByteArray stringData(stringSize * 2);
    file.readBlock(stringData.data(), stringSize * 2);

    // decode the raw data into a string
    QTextIStream stringStream(stringData);
    stringStream.setEncoding(QTextStream::RawUnicode);
    QString string = stringStream.read();

    // strip any \0 chars
    int i = string.find((char) 0);
    if (i >= 0) string.truncate(i);

    // done
    return string;
}
*/
