 /*****************************************************************************
 *   Copyright (C) 2004 by Andrew de Quincey                                  *
 *   adq@lidskialf.net                                                        *
 *   Copyright (C) 2005 by Michael Schulze                                    *
 *   mike.s@genion.de                                                         *
 *                                                                            *
 *  The code contained in this file is free software; you can redistribute    *
 *  it and/or modify it under the terms of the GNU Lesser General Public      *
 *  License as published by the Free Software Foundation; either version      *
 *  2.1 of the License, or (at your option) any later version.                *
 *                                                                            *
 *  This file is distributed in the hope that it will be useful,              *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 *  Lesser General Public License for more details.                           *
 *                                                                            *
 *  You should have received a copy of the GNU Lesser General Public          *
 *  License along with this code; if not, write to the Free Software          *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *                                                                            *
 *  iTunes and iPod are trademarks of Apple                                   *
 *                                                                            *
 *  This product is not supported/written/published by Apple!                 *
 *****************************************************************************/

#include "ipodsysinfo.h"

#include <qfile.h>
#include <qregexp.h>
#include <qtextstream.h>

#ifdef WIN32
#include <Windows.h>
#else
#include <sys/statfs.h>
#endif

IPodSysInfo::IPodSysInfo( const QString& ipod_Base )
    : m_ipodBase( ipod_Base ), /* m_deviceName( devicename ), */m_musicDir( m_ipodBase ),
      m_numTrackFileDirs( 0 ), m_kbytesTotal(0), m_kbytesAvail(0)
{
    if ( ! m_ipodBase.endsWith( "/" ) ) {
        m_ipodBase.append( '/' );
    }
}


IPodSysInfo::~IPodSysInfo() {}


QDir IPodSysInfo::findMusicDir( const QString& ipodbase, bool * found ) const {
    QDir dir( ipodbase );
    dir.cd( IPOD_CONTROLDIR );

    QStringList musicDirNames = dir.entryList( "Music;music", QDir::Dirs | QDir::Hidden );
    if ( musicDirNames.size() < 1 ) {
        // Dunno where to look for music dirs
        if ( found != NULL ) {
            *found = false;
        }
        return dir;
    }

    for ( unsigned int i = 0; i < musicDirNames.size(); ++i ) {
        if ( dir.cd ( musicDirNames[ i ] ) ) {
            dir.setFilter( QDir::Dirs | QDir::Hidden );
            dir.setNameFilter( "F??;f??" );
            if ( dir.count() > 0 ) {
                if ( found != NULL ) {
                    *found = true;
                }
                return dir;
            } else {
                dir.cdUp();
            }
        }
    }

    if ( found != NULL ) {
        *found = false;
    }

    return dir;
}


bool IPodSysInfo::createMusicDirs() {
    bool musicDirFound = false;
    findMusicDir( m_ipodBase, &musicDirFound );
    if ( musicDirFound ) {
        // could not create - musicdir already exists
        return false;
    }

    // not found - create musicdirs
    QString dirname( IPOD_MUSICDIR );
    QDir dir( getIPodControlDirName() );

    dir.mkdir( dirname );
    if ( !dir.cd( dirname ) ) {
        return false;
    }

    uint numMusicDirs = getTotalDiskSpaceKB() / 1048576 < 5 ? 20 : 50;
    for ( uint i = 0; i < numMusicDirs; i++ ) {
        dir.mkdir( QString().sprintf("f%02d", i ) );
    }

    return true;
}


QString IPodSysInfo::findNewTrackFileName( Q_UINT32 id ) {
    QString dirName = m_musicDir[ id % m_numTrackFileDirs ];

    // try to find an unused filename with "kpod<id>.*"
    QDir directory = m_musicDir;
    directory.cd( dirName );
    directory.setFilter( QDir::Files );

    QString subIDstr( "" );
    int subID = 0;
    do {
        if ( subID ) {
            subIDstr = QString( "_" ); 
            subIDstr.append( QString::number( subID ) );
        }
    } while ( directory.entryList( "kpod" + QString::number( id ) + subIDstr + ".*" ).count() > 0 && ++subID );

    // now we have an unused filename
    // ...
    // so form the trackpath

    QString trackpath;
    trackpath.sprintf( ":iPod_Control:Music:%s:%s", dirName.ascii(),
        ( "kpod" + QString::number( id )  + subIDstr ).ascii());

    return trackpath;
}


bool IPodSysInfo::load() {
    bool result = true;

    // load the sysinfo file first - uninitialized ipods may only have these
    QFile sysInfoFile( m_ipodBase + IPOD_CONTROLDIR + "/Device/SysInfo" );
    if ( sysInfoFile.exists() && sysInfoFile.open( IO_ReadOnly ) ) {
        QRegExp re( "^(.+):(.+)$" );
        QTextStream stringStream( &sysInfoFile );
        stringStream.setEncoding( QTextStream::UnicodeUTF8 ); // guess

        while( !stringStream.atEnd() ) {
        QString line = stringStream.readLine();
            if ( !re.search( line ) ) {
                QString key = re.cap( 1 ).stripWhiteSpace();
                QString value = re.cap( 2 ).stripWhiteSpace();
                if ( key.length() ) {
                    m_details[ key ] = value;
                }
            }
        }

        sysInfoFile.close();
        result &= true;
    }   // don't care if this file is missing

    // refresh disk usage statistics
    refreshDiskUsageStats();

    // Try to figure out where the music files are
    bool musicDirFound = false;
    m_musicDir = findMusicDir( m_ipodBase, &musicDirFound );
    m_numTrackFileDirs = m_musicDir.count();

    result &= musicDirFound;

    return result;
}


IPodSysInfo::KeyIterator IPodSysInfo::getKeys() const {
    return KeyIterator( m_details.begin(), m_details.end() );
}

QString IPodSysInfo::getValue(const QString& key) const {
    PropertyMap::const_iterator elem = m_details.find( key );
    if ( elem == m_details.end() ) {
        return itunesdb::utils::NullQString;
    }
    return (*elem).second;
}

void IPodSysInfo::refreshDiskUsageStats() {
    this->m_kbytesTotal = 0;
}

unsigned long IPodSysInfo::getTotalDiskSpaceKB() {
    ensureDiskUsageStats();
    return m_kbytesTotal;
}

unsigned long IPodSysInfo::getAvailableDiskSpaceKB() {
    ensureDiskUsageStats();
    return m_kbytesAvail;
}

unsigned long IPodSysInfo::getUsedDiskSpaceKB() {
    ensureDiskUsageStats();
    return m_kbytesTotal - m_kbytesAvail;
}

void IPodSysInfo::ensureDiskUsageStats() {
    if ( ! this->m_kbytesTotal ) {
#ifdef WIN32
        ULARGE_INTEGER bytesavail;
        ULARGE_INTEGER bytestotal;
        GetDiskFreeSpaceExA( m_ipodBase.ascii(), &bytesavail, &bytestotal, NULL );
        this->m_kbytesAvail = bytesavail.QuadPart >> 10;
        this->m_kbytesTotal = bytestotal.QuadPart >> 10;
#else
        struct statfs statistics;
        if ( statfs( ( m_ipodBase + IPOD_CONTROLDIR ).ascii(), &statistics ) == 0 ) {
            this->m_kbytesTotal = long( statistics.f_bsize * ( statistics.f_blocks >> 10 ) );
            this->m_kbytesAvail = long( statistics.f_bsize * ( statistics.f_bavail >> 10 ) );
        }
#endif
    }
}

void IPodSysInfo::clear() {
    refreshDiskUsageStats();
    m_details.clear();
    m_ipodBase = QString();
//    m_deviceName = QString();
    m_musicDir = QDir();
}

