 /*****************************************************************************
 *   Copyright (C) 2006 by Michael Schulze                                    *
 *   mike.s@genion.de                                                         *
 *                                                                            *
 *  The code contained in this file is free software; you can redistribute    *
 *  it and/or modify it under the terms of the GNU Lesser General Public      *
 *  License as published by the Free Software Foundation; either version      *
 *  2.1 of the License, or (at your option) any later version.                *
 *                                                                            *
 *  This file is distributed in the hope that it will be useful,              *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 *  Lesser General Public License for more details.                           *
 *                                                                            *
 *  You should have received a copy of the GNU Lesser General Public          *
 *  License along with this code; if not, write to the Free Software          *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *                                                                            *
 *  iTunes and iPod are trademarks of Apple                                   *
 *                                                                            *
 *  This product is not supported/written/published by Apple!                 *
 *****************************************************************************/


#include "itunesdb.h"

using namespace itunesdb;

/**
 * Represents a string based playlist action
 */
class StringPredicate {
public:
    virtual ~StringPredicate() {}
    virtual SPLStringAction getAction() const = 0;
    virtual bool operator() ( const QString& toCheck, const QString& value ) const = 0;
};

/**
 * Represents a uint based playlist action
 */
class UIntPredicate {
public:
    virtual ~UIntPredicate() {}
    virtual SPLUIntAction getAction() const = 0;
    virtual bool operator() ( Q_UINT64 toCheck, Q_UINT64 from, Q_UINT64 to ) const = 0;
};

/**
 * Accessor to String based properties of tracks
 */
class StringPropertyAccessor {
public:
    virtual ~StringPropertyAccessor() {}
    virtual SPLStringField getField() const = 0;
    virtual const QString& operator() ( const itunesdb::Track& track ) const = 0;
};

/**
 * Accessor to uint based properties of tracks
 */
class UIntPropertyAccessor {
public:
    virtual ~UIntPropertyAccessor() {}
    virtual SPLUIntField getField() const = 0;
    virtual Q_UINT64 operator() ( const itunesdb::Track& track ) const = 0;
};




/***
 *
 * Accessors
 *
 ***/


class TrackPropertyAccessors {

    friend class ITunesDBSPLRuleSet;

    static StringPropertyAccessor * getStringAccessorFor( Q_UINT32 field ) {
        switch ( field ) {
        case FIELD_NAME:
            return &TITLE;
        case FIELD_ALBUM:
            return &ALBUM;
        case FIELD_ARTIST:
            return &ARTIST;
        case FIELD_GENRE:
            return &GENRE;
        case FIELD_KIND:
            return &KIND;
        case FIELD_COMMENT:
            return &COMMENT;
        case FIELD_COMPOSER:
            return &COMPOSER;
        case FIELD_GROUPING:
            return &GROUPING;
        case FIELD_DESCRIPTION:
            return &DESCRIPTION;
        case FIELD_CATEGORY:
            return &CATEGORY;
        case FIELD_TV_SHOW:
            return &TVSHOW;
        case FIELD_ALBUM_ARTIST:
            return &ALBUMARTIST;
        default:
            return NULL;
        }
    }

    static UIntPropertyAccessor * getUIntAccessorFor( Q_UINT32 field ) {
        switch ( field ) {
        case FIELD_BITRATE:
            return &BITRATE;
        case FIELD_SAMPLERATE:
            return &SAMPLERATE;
        case FIELD_YEAR:
            return &YEAR;
        case FIELD_DATEMODIFIED:
            return &DATE_MODIFIED;
        case FIELD_TRACKNUMBER:
            return &TRACKNUMBER;
        case FIELD_SIZE:
            return &SIZE;
        case FIELD_TIME:
            return &TIME;
        case FIELD_DATEADDED:
            return &DATEADDED;
        case FIELD_PLAYCOUNT:
            return &PLAYCOUNT;
        case FIELD_LASTPLAYED:
            return &LASTPLAYED;
        case FIELD_DISCNUMBER:
            return &DISCNUMBER;
        case FIELD_RATING:
            return &RATING;
        case FIELD_COMPILATION:
            return &COMPILATIONFLAG;
        case FIELD_BPM:
            return &BEATSPERMINUTE;
        case FIELD_VIDEO_KIND:
            return &VIDEO_KIND;
        case FIELD_SEASON_NR:
            return &SEASON;
        case FIELD_SKIP_COUNT:
            return &SKIP_COUNT;
        case FIELD_LAST_SKIPPED:
            return &LAST_SKIPPED;
        case FIELD_PODCAST:   // TODO implement
        default:
            return NULL;
        }
    }

    static class GetArtist : public StringPropertyAccessor {
        virtual SPLStringField getField() const {
            return FIELD_ARTIST;
        }
        virtual const QString& operator() ( const Track& track ) const {
            return track.getArtist();
        }
    } ARTIST;

    static class GetAlbum : public StringPropertyAccessor {
        virtual SPLStringField getField() const {
            return FIELD_ALBUM;
        }
        virtual const QString& operator() ( const Track& track ) const {
            return track.getAlbum();
        }
    } ALBUM;

    static class GetTitle : public StringPropertyAccessor {
        virtual SPLStringField getField() const {
            return FIELD_NAME;
        }
        virtual const QString& operator() ( const Track& track ) const {
            return track.getTitle();
        }
    } TITLE;

    static class GetGenre : public StringPropertyAccessor {
        virtual SPLStringField getField() const {
            return FIELD_GENRE;
        }
        virtual const QString& operator() ( const Track& track ) const {
            return track.getGenre();
        }
    } GENRE;

    static class GetKind : public StringPropertyAccessor {
        virtual SPLStringField getField() const {
            return FIELD_KIND;
        }
        virtual const QString& operator() ( const Track& track ) const {
            return track.getFileFormatDesc();
        }
    } KIND;

    static class GetComment : public StringPropertyAccessor {
        virtual SPLStringField getField() const {
            return FIELD_COMMENT;
        }
        virtual const QString& operator() ( const Track& track ) const {
            return track.getComment();
        }
    } COMMENT;

    static class GetComposer : public StringPropertyAccessor {
        virtual SPLStringField getField() const {
            return FIELD_COMPOSER;
        }
        virtual const QString& operator() ( const Track& track ) const {
            return track.getComposer();
        }
    } COMPOSER;

    static class GetGrouping : public StringPropertyAccessor {
        virtual SPLStringField getField() const {
            return FIELD_GROUPING;
        }
        virtual const QString& operator() ( const Track& track ) const {
            return track.getGrouping();
        }
    } GROUPING;

    static class GetDescription : public StringPropertyAccessor {
        virtual SPLStringField getField() const {
            return FIELD_DESCRIPTION;
        }
        virtual const QString& operator() ( const Track& track ) const {
            return track.getDescription();
        }
    } DESCRIPTION;

    static class GetCategory : public StringPropertyAccessor {
        virtual SPLStringField getField() const {
            return FIELD_CATEGORY;
        }
        virtual const QString& operator() ( const Track& track ) const {
            return track.getCategory();
        }
    } CATEGORY;

    static class GetTVShow : public StringPropertyAccessor {
        virtual SPLStringField getField() const {
            return FIELD_TV_SHOW;
        }
        virtual const QString& operator() ( const Track& track ) const {
            return track.getTvShow();
        }
    } TVSHOW;

    static class GetAlbumArtist : public StringPropertyAccessor {
        virtual SPLStringField getField() const {
            return FIELD_ALBUM_ARTIST;
        }
        virtual const QString& operator() ( const Track& track ) const {
            return track.getAlbumArtist();
        }
    } ALBUMARTIST;


    /******************************************************
     *                                                    *
     *                 INTEGER BASED FIELDS               *
     *                                                    *
     ******************************************************/


    static class GetBitrate : public UIntPropertyAccessor {
        virtual SPLUIntField getField() const {
            return FIELD_BITRATE;
        }
        virtual Q_UINT64 operator() ( const Track& track ) const {
            return track.getBitrate();
        }
    } BITRATE;

    static class GetSamplerate : public UIntPropertyAccessor {
        virtual SPLUIntField getField() const {
            return FIELD_SAMPLERATE;
        }
        virtual Q_UINT64 operator() ( const Track& track ) const {
            return track.getSamplerate();
        }
    } SAMPLERATE;

    static class GetYear : public UIntPropertyAccessor {
        virtual SPLUIntField getField() const {
            return FIELD_YEAR;
        }
        virtual Q_UINT64 operator() ( const Track& track ) const {
            return track.getYear();
        }
    } YEAR;

    static class GetDateModified : public UIntPropertyAccessor {
        virtual SPLUIntField getField() const {
            return FIELD_DATEMODIFIED;
        }
        virtual Q_UINT64 operator() ( const Track& track ) const {
            return track.getLastModified();
        }
    } DATE_MODIFIED;

    static class GetTrackNumber : public UIntPropertyAccessor {
        virtual SPLUIntField getField() const {
            return FIELD_TRACKNUMBER;
        }
        virtual Q_UINT64 operator() ( const Track& track ) const {
            return track.getTrackNumber();
        }
    } TRACKNUMBER;

    static class GetSize : public UIntPropertyAccessor {
        virtual SPLUIntField getField() const {
            return FIELD_SIZE;
        }
        virtual Q_UINT64 operator() ( const Track& track ) const {
            return track.getFileSize();
        }
    } SIZE;

    static class GetTime : public UIntPropertyAccessor {
        virtual SPLUIntField getField() const {
            return FIELD_TIME;
        }
        virtual Q_UINT64 operator() ( const Track& track ) const {
            return track.getTrackLength();
        }
    } TIME;

    static class GetDateAdded : public UIntPropertyAccessor {
        virtual SPLUIntField getField() const {
            return FIELD_DATEADDED;
        }
        virtual Q_UINT64 operator() ( const Track& track ) const {
            return track.getDateAdded();
        }
    } DATEADDED;

    static class GetPlayCount : public UIntPropertyAccessor {
        virtual SPLUIntField getField() const {
            return FIELD_PLAYCOUNT;
        }
        virtual Q_UINT64 operator() ( const Track& track ) const {
            return track.getPlayCount();
        }
    } PLAYCOUNT;

    static class GetLastPlayed : public UIntPropertyAccessor {
        virtual SPLUIntField getField() const {
            return FIELD_LASTPLAYED;
        }
        virtual Q_UINT64 operator() ( const Track& track ) const {
            return track.getLastPlayed();
        }
    } LASTPLAYED;

    static class GetDiscNumber : public UIntPropertyAccessor {
        virtual SPLUIntField getField() const {
            return FIELD_DISCNUMBER;
        }
        virtual Q_UINT64 operator() ( const Track& track ) const {
            return track.getCdNumber();
        }
    } DISCNUMBER;

    static class GetRating : public UIntPropertyAccessor {
        virtual SPLUIntField getField() const {
            return FIELD_RATING;
        }
        virtual Q_UINT64 operator() ( const Track& track ) const {
            return track.getRating();
        }
    } RATING;

    static class GetCompilationFlag : public UIntPropertyAccessor {
        virtual SPLUIntField getField() const {
            return FIELD_COMPILATION;
        }
        virtual Q_UINT64 operator() ( const Track& track ) const {
            return track.getCompilationFlag();
        }
    } COMPILATIONFLAG;

    static class GetBPM : public UIntPropertyAccessor {
        virtual SPLUIntField getField() const {
            return FIELD_BPM;
        }
        virtual Q_UINT64 operator() ( const Track& track ) const {
            return track.getBPM();
        }
    } BEATSPERMINUTE;

    static class GetVideoKind : public UIntPropertyAccessor {
        virtual SPLUIntField getField() const {
            return FIELD_VIDEO_KIND;
        }
        virtual Q_UINT64 operator() ( const Track& track ) const {
            return track.isVideo();
        }
    } VIDEO_KIND;

    static class GetSeasonNumber : public UIntPropertyAccessor {
        virtual SPLUIntField getField() const {
            return FIELD_SEASON_NR;
        }
        virtual Q_UINT64 operator() ( const Track& track ) const {
            return track.getSeasonNumber();
        }
    } SEASON;

    static class GetSkipCount : public UIntPropertyAccessor {
        virtual SPLUIntField getField() const {
            return FIELD_SKIP_COUNT;
        }
        virtual Q_UINT64 operator() ( const Track& track ) const {
            return track.getSkipCount();
        }
    } SKIP_COUNT;

    static class GetLastSkipped : public UIntPropertyAccessor {
        virtual SPLUIntField getField() const {
            return FIELD_LAST_SKIPPED;
        }
        virtual Q_UINT64 operator() ( const Track& track ) const {
            return track.getLastSkippedTime();
        }
    } LAST_SKIPPED;
};

TrackPropertyAccessors::GetArtist TrackPropertyAccessors::ARTIST =
    TrackPropertyAccessors::GetArtist();
TrackPropertyAccessors::GetAlbum TrackPropertyAccessors::ALBUM =
    TrackPropertyAccessors::GetAlbum();
TrackPropertyAccessors::GetTitle TrackPropertyAccessors::TITLE =
    TrackPropertyAccessors::GetTitle();
TrackPropertyAccessors::GetGenre TrackPropertyAccessors::GENRE =
    TrackPropertyAccessors::GetGenre();
TrackPropertyAccessors::GetKind TrackPropertyAccessors::KIND =
    TrackPropertyAccessors::GetKind();
TrackPropertyAccessors::GetComment TrackPropertyAccessors::COMMENT =
    TrackPropertyAccessors::GetComment();
TrackPropertyAccessors::GetComposer TrackPropertyAccessors::COMPOSER =
    TrackPropertyAccessors::GetComposer();
TrackPropertyAccessors::GetGrouping TrackPropertyAccessors::GROUPING =
    TrackPropertyAccessors::GetGrouping();
TrackPropertyAccessors::GetDescription TrackPropertyAccessors::DESCRIPTION =
    TrackPropertyAccessors::GetDescription();
TrackPropertyAccessors::GetCategory TrackPropertyAccessors::CATEGORY =
    TrackPropertyAccessors::GetCategory();
TrackPropertyAccessors::GetTVShow TrackPropertyAccessors::TVSHOW =
    TrackPropertyAccessors::GetTVShow();
TrackPropertyAccessors::GetAlbumArtist TrackPropertyAccessors::ALBUMARTIST =
    TrackPropertyAccessors::GetAlbumArtist();

TrackPropertyAccessors::GetBitrate TrackPropertyAccessors::BITRATE =
    TrackPropertyAccessors::GetBitrate();
TrackPropertyAccessors::GetSamplerate TrackPropertyAccessors::SAMPLERATE =
    TrackPropertyAccessors::GetSamplerate();
TrackPropertyAccessors::GetYear TrackPropertyAccessors::YEAR =
    TrackPropertyAccessors::GetYear();
TrackPropertyAccessors::GetDateModified TrackPropertyAccessors::DATE_MODIFIED =
    TrackPropertyAccessors::GetDateModified();
TrackPropertyAccessors::GetTrackNumber TrackPropertyAccessors::TRACKNUMBER =
    TrackPropertyAccessors::GetTrackNumber();
TrackPropertyAccessors::GetSize TrackPropertyAccessors::SIZE =
    TrackPropertyAccessors::GetSize();
TrackPropertyAccessors::GetTime TrackPropertyAccessors::TIME =
    TrackPropertyAccessors::GetTime();
TrackPropertyAccessors::GetDateAdded TrackPropertyAccessors::DATEADDED =
    TrackPropertyAccessors::GetDateAdded();
TrackPropertyAccessors::GetPlayCount TrackPropertyAccessors::PLAYCOUNT =
    TrackPropertyAccessors::GetPlayCount();
TrackPropertyAccessors::GetLastPlayed TrackPropertyAccessors::LASTPLAYED =
    TrackPropertyAccessors::GetLastPlayed();
TrackPropertyAccessors::GetDiscNumber TrackPropertyAccessors::DISCNUMBER =
    TrackPropertyAccessors::GetDiscNumber();
TrackPropertyAccessors::GetRating TrackPropertyAccessors::RATING =
    TrackPropertyAccessors::GetRating();
TrackPropertyAccessors::GetCompilationFlag TrackPropertyAccessors::COMPILATIONFLAG =
    TrackPropertyAccessors::GetCompilationFlag();
TrackPropertyAccessors::GetBPM TrackPropertyAccessors::BEATSPERMINUTE =
    TrackPropertyAccessors::GetBPM();
TrackPropertyAccessors::GetVideoKind TrackPropertyAccessors::VIDEO_KIND =
    TrackPropertyAccessors::GetVideoKind();
TrackPropertyAccessors::GetSeasonNumber TrackPropertyAccessors::SEASON =
    TrackPropertyAccessors::GetSeasonNumber();
TrackPropertyAccessors::GetSkipCount TrackPropertyAccessors::SKIP_COUNT =
    TrackPropertyAccessors::GetSkipCount();
TrackPropertyAccessors::GetLastSkipped TrackPropertyAccessors::LAST_SKIPPED =
    TrackPropertyAccessors::GetLastSkipped();


/***
 *
 * SPLPredicates
 *
 ***/


class SPLPredicates {

    friend class ITunesDBSPLRuleSet;

    static const StringPredicate * getStringPredicateFor( Q_UINT32 action ) {
        switch ( action ) {
        case ACTION_IS_STRING:
            return &IS_STRING;
        case ACTION_CONTAINS:
            return &CONTAINS;
        case ACTION_STARTSWITH:
            return &STARTSWITH;
        case ACTION_ENDSWITH:
            return &ENDSWITH;
        case ACTION_IS_NOT_STRING:
            return &IS_NOT_STRING;
        case ACTION_DOESNT_CONTAIN:
            return &DOESNT_CONTAIN;
        case ACTION_DOESNT_STARTWITH:
            return &DOESNT_STARTWITH;
        case ACTION_DOESNT_ENDWITH:
            return &DOESNT_ENDWITH;
        default:
            return NULL;
        }
    }

    static const UIntPredicate * getUIntPredicateFor( Q_UINT32 action ) {
        switch ( action ) {
        case ACTION_IS_UINT:
            return &IS_UINT;
        case ACTION_GREATER_THAN:
            return &GREATER_THAN;
        case ACTION_GREATER_OR_EQUAL:
            return &GREATER_OR_EQUAL;
        case ACTION_LESS_THAN:
            return &LESS_THAN;
        case ACTION_LESS_OR_EQUAL:
            return &LESS_OR_EQUAL;
        case ACTION_IN_RANGE:
            return &IN_RANGE;
        case ACTION_IN_THE_LAST:
            return &IN_THE_LAST;
        case ACTION_IS_NOT_UINT:
            return &NOT_IS_UINT;
        case ACTION_NOT_GREATER:
            return &NOT_GREATER;
        case ACTION_NOT_GREATER_OR_EQUAL:
            return &NOT_GREATER_OR_EQUAL;
        case ACTION_NOT_LESS:
            return &NOT_LESS;
        case ACTION_NOT_LESS_OR_EQUAL:
            return &NOT_LESS_OR_EQUAL;
        case ACTION_NOT_IN_RANGE:
            return &NOT_IN_RANGE;
        case ACTION_NOT_IN_THE_LAST:
            return &NOT_IN_THE_LAST;
        default:
            return NULL;
        }
    }

    static class IsString : public StringPredicate {
        virtual SPLStringAction getAction() const {
            return ACTION_IS_STRING;
        }
        virtual bool operator() ( const QString& toCheck, const QString& testString ) const {
            return toCheck.lower().compare( testString.lower() ) == 0;
        }
    } IS_STRING;

    static class Contains : public StringPredicate {
        virtual SPLStringAction getAction() const {
            return ACTION_CONTAINS;
        }
        virtual bool operator() ( const QString& toCheck, const QString& testString ) const {
            return toCheck.find( testString, 0, false ) != -1;
        }
    } CONTAINS;

    static class StartsWith : public StringPredicate {
        virtual SPLStringAction getAction() const {
            return ACTION_STARTSWITH;
        }
        virtual bool operator() ( const QString& toCheck, const QString& testString ) const {
            // "e" - das war Claudia
            return toCheck.startsWith( testString, false );
        }
    } STARTSWITH;

    static class EndsWith : public StringPredicate {
        virtual SPLStringAction getAction() const {
            return ACTION_ENDSWITH;
        }
        virtual bool operator() ( const QString& toCheck, const QString& testString ) const {
            return toCheck.endsWith( testString, false );
        }
    } ENDSWITH;

    static class IsUInt : public UIntPredicate {
        virtual SPLUIntAction getAction() const {
            return ACTION_IS_UINT;
        }
        virtual bool operator() ( Q_UINT64 toCheck, Q_UINT64 from, Q_UINT64 to ) const {
            return toCheck == from || toCheck == to ;
        }
    } IS_UINT;

    static class GreaterThan : public UIntPredicate {
        virtual SPLUIntAction getAction() const {
            return ACTION_GREATER_THAN;
        }
        virtual bool operator() ( Q_UINT64 toCheck, Q_UINT64 from, Q_UINT64 to ) const {
            return toCheck > from && toCheck > to;
        }
    } GREATER_THAN;

    static class GreaterOrEqual : public UIntPredicate {
        virtual SPLUIntAction getAction() const {
            return ACTION_GREATER_OR_EQUAL;
        }
        virtual bool operator() ( Q_UINT64 toCheck, Q_UINT64 from, Q_UINT64 to ) const {
            return toCheck >= from && toCheck >= to;
        }
    } GREATER_OR_EQUAL;

    static class LessThan : public UIntPredicate {
        virtual SPLUIntAction getAction() const {
            return ACTION_LESS_THAN;
        }
        virtual bool operator() ( Q_UINT64 toCheck, Q_UINT64 from, Q_UINT64 to ) const {
            return toCheck < from || toCheck < to;
        }
    } LESS_THAN;

    static class LessOrEqual : public UIntPredicate {
        virtual SPLUIntAction getAction() const {
            return ACTION_LESS_OR_EQUAL;
        }
        virtual bool operator() ( Q_UINT64 toCheck, Q_UINT64 from, Q_UINT64 to ) const {
            return toCheck <= from || toCheck <= to;
        }
    } LESS_OR_EQUAL;

    static class InTheLast : public UIntPredicate {
        virtual SPLUIntAction getAction() const {
            return ACTION_IN_THE_LAST;
        }
        virtual bool operator() ( Q_UINT64 toCheck, Q_UINT64 from, Q_UINT64 to ) const {
            return toCheck >= from && toCheck <= to;
        }
    } IN_THE_LAST;

    static class InRange : public UIntPredicate {
        virtual SPLUIntAction getAction() const {
            return ACTION_IN_RANGE;
        }
        virtual bool operator() ( Q_UINT64 toCheck, Q_UINT64 from, Q_UINT64 to ) const {
            return toCheck >= from && toCheck <= to;
        }
    } IN_RANGE;

    static class UIntPredicateNegator : public UIntPredicate {
        const UIntPredicate& mToNegate;
    public:
        UIntPredicateNegator( const UIntPredicate& toNegate )
            : mToNegate( toNegate )
        {}
        virtual ~UIntPredicateNegator() {}
        virtual SPLUIntAction getAction() const {
            switch ( mToNegate.getAction() ) {
            case ACTION_IS_UINT:
                return ACTION_IS_NOT_UINT;
            case ACTION_GREATER_THAN:
                return ACTION_NOT_GREATER;
            case ACTION_GREATER_OR_EQUAL:
                return ACTION_NOT_GREATER_OR_EQUAL;
            case ACTION_LESS_THAN:
                return ACTION_NOT_LESS;
            case ACTION_LESS_OR_EQUAL:
                return ACTION_NOT_LESS_OR_EQUAL;
            case ACTION_IN_RANGE:
                return ACTION_NOT_IN_RANGE;
            case ACTION_IN_THE_LAST:
                return ACTION_NOT_IN_THE_LAST;
            default:
                return ACTION_NOT_IN_RANGE;
            }
        }
        virtual bool operator() ( Q_UINT64 toCheck, Q_UINT64 from, Q_UINT64 to ) const {
            return !mToNegate( toCheck, from, to );
        }
    } NOT_IS_UINT, NOT_GREATER, NOT_GREATER_OR_EQUAL, NOT_LESS, NOT_LESS_OR_EQUAL, NOT_IN_RANGE, NOT_IN_THE_LAST;

    static class StringPredicateNegator : public StringPredicate {
        const StringPredicate& mToNegate;
    public:
        StringPredicateNegator( const StringPredicate& toNegate )
            : mToNegate( toNegate )
        {}
        virtual ~StringPredicateNegator() {}
        virtual SPLStringAction getAction() const {
            switch ( mToNegate.getAction() ) {
            case ACTION_IS_STRING:
                return ACTION_IS_NOT_STRING;
            case ACTION_CONTAINS:
                return ACTION_DOESNT_CONTAIN;
            case ACTION_STARTSWITH:
                return ACTION_DOESNT_STARTWITH;
            case ACTION_ENDSWITH:
                return ACTION_DOESNT_ENDWITH;
            default:
                return ACTION_IS_NOT_STRING;
            }
        }
        virtual bool operator() ( const QString& toCheck, const QString& testString ) const {
            return !mToNegate( toCheck, testString );
        }
    } IS_NOT_STRING, DOESNT_CONTAIN, DOESNT_STARTWITH, DOESNT_ENDWITH;

};


SPLPredicates::IsString SPLPredicates::IS_STRING =
    SPLPredicates::IsString();
SPLPredicates::Contains SPLPredicates::CONTAINS =
    SPLPredicates::Contains();
SPLPredicates::StartsWith SPLPredicates::STARTSWITH =
    SPLPredicates::StartsWith();
SPLPredicates::EndsWith SPLPredicates::ENDSWITH =
    SPLPredicates::EndsWith();
SPLPredicates::StringPredicateNegator SPLPredicates::IS_NOT_STRING =
    SPLPredicates::StringPredicateNegator( SPLPredicates::IS_STRING );
SPLPredicates::StringPredicateNegator SPLPredicates::DOESNT_CONTAIN =
    SPLPredicates::StringPredicateNegator( SPLPredicates::CONTAINS );
SPLPredicates::StringPredicateNegator SPLPredicates::DOESNT_STARTWITH =
    SPLPredicates::StringPredicateNegator( SPLPredicates::STARTSWITH );
SPLPredicates::StringPredicateNegator SPLPredicates::DOESNT_ENDWITH =
    SPLPredicates::StringPredicateNegator( SPLPredicates::ENDSWITH );


SPLPredicates::IsUInt SPLPredicates::IS_UINT =
    SPLPredicates::IsUInt();
SPLPredicates::GreaterThan SPLPredicates::GREATER_THAN =
    SPLPredicates::GreaterThan();
SPLPredicates::GreaterOrEqual SPLPredicates::GREATER_OR_EQUAL =
    SPLPredicates::GreaterOrEqual();
SPLPredicates::LessThan SPLPredicates::LESS_THAN =
    SPLPredicates::LessThan();
SPLPredicates::LessOrEqual SPLPredicates::LESS_OR_EQUAL =
    SPLPredicates::LessOrEqual();
SPLPredicates::InRange SPLPredicates::IN_RANGE =
    SPLPredicates::InRange();
SPLPredicates::InTheLast SPLPredicates::IN_THE_LAST =
    SPLPredicates::InTheLast();
SPLPredicates::UIntPredicateNegator SPLPredicates::NOT_IS_UINT =
    SPLPredicates::UIntPredicateNegator( SPLPredicates::IS_UINT );
SPLPredicates::UIntPredicateNegator SPLPredicates::NOT_GREATER =
    SPLPredicates::UIntPredicateNegator( SPLPredicates::GREATER_THAN );
SPLPredicates::UIntPredicateNegator SPLPredicates::NOT_GREATER_OR_EQUAL =
    SPLPredicates::UIntPredicateNegator( SPLPredicates::GREATER_OR_EQUAL );
SPLPredicates::UIntPredicateNegator SPLPredicates::NOT_LESS =
    SPLPredicates::UIntPredicateNegator( SPLPredicates::LESS_THAN );
SPLPredicates::UIntPredicateNegator SPLPredicates::NOT_LESS_OR_EQUAL =
    SPLPredicates::UIntPredicateNegator( SPLPredicates::LESS_OR_EQUAL );
SPLPredicates::UIntPredicateNegator SPLPredicates::NOT_IN_RANGE =
    SPLPredicates::UIntPredicateNegator( SPLPredicates::IN_RANGE );
SPLPredicates::UIntPredicateNegator SPLPredicates::NOT_IN_THE_LAST =
    SPLPredicates::UIntPredicateNegator( SPLPredicates::IN_THE_LAST );


/***
 *
 * SPLStringTrackPredicate
 *
 ***/


class SPLStringTrackPredicate : public SmartPlaylistRule, public TrackPredicate {
    const StringPredicate& mStringPredicate;
    const StringPropertyAccessor& mStringPropertyAccessor;

public:

    SPLStringTrackPredicate( const StringPredicate& predicate,
                             const StringPropertyAccessor& accessor,
                             const QString& toCheckAgainst )
        : SmartPlaylistRule( accessor.getField(), predicate.getAction(), toCheckAgainst ),
          mStringPredicate( predicate ), mStringPropertyAccessor( accessor )
    {
    }

    virtual ~SPLStringTrackPredicate()
    {
    }

    virtual bool operator() ( const Track * track ) const {
        return track && mStringPredicate( mStringPropertyAccessor( *track ), getString() );
    }

};


/***
 *
 * SPLUIntTrackPredicate
 *
 ***/


class SPLUIntTrackPredicate : public SmartPlaylistRule, public TrackPredicate {
    const UIntPredicate& mUIntPredicate;
    const UIntPropertyAccessor& mUIntPropertyAccessor;

public:

    SPLUIntTrackPredicate( const UIntPredicate& predicate,
                           const UIntPropertyAccessor& accessor,
                           Q_UINT64 fromValue, Q_UINT64 toValue,
                           Q_UINT64 fromDate, Q_UINT64 toDate,
                           Q_UINT64 fromUnits, Q_UINT64 toUnits )
        : SmartPlaylistRule( accessor.getField(), predicate.getAction(),
                             fromValue, toValue, fromDate, toDate, fromUnits, toUnits ),
          mUIntPredicate( predicate ), mUIntPropertyAccessor( accessor )
    {
    }

    virtual ~SPLUIntTrackPredicate()
    {
    }

    virtual bool operator() ( const Track * track ) const {
        return track && mUIntPredicate( mUIntPropertyAccessor( *track ), calculatedFrom(), calculatedTo() );
    }

};


/***
 *
 * SPLPlaylistIDPredicate
 *
 ***/


class SPLTrackInPlaylistPredicate : public SmartPlaylistRule, public TrackPredicate {

    const Playlist * mPlaylist;
    const bool mNegate;

public:

    SPLTrackInPlaylistPredicate( const Playlist * playlist,
                                 SPLUIntAction action,
                                 Q_UINT64 fromValue, Q_UINT64 toValue,
                                 Q_UINT64 fromDate, Q_UINT64 toDate,
                                 Q_UINT64 fromUnits, Q_UINT64 toUnits )
        : SmartPlaylistRule( FIELD_PLAYLISTID, action, fromValue, toValue,
                             fromDate, toDate, fromUnits, toUnits ),
          mPlaylist( playlist ),
          mNegate( action & ACTION_NOT_BIT )
    {
    }

    virtual ~SPLTrackInPlaylistPredicate()
    {
    }

    virtual bool operator() ( const Track * track ) const {
        return track && ( mNegate ^ mPlaylist->contains( track->getID() ) );
    }

};


/***
 *
 * ITunesDBSPLRuleSet Impl
 *
 ***/


ITunesDBSPLRuleSet::ITunesDBSPLRuleSet( ITunesDBPlaylist& playlist, const ITunesDB& itunesdb )
    : SmartPlaylistRuleSet(), mItunesdb( itunesdb ),
      mPlaylist( playlist ), mLastSeenItunesDBVersion( 0 ),
      mLastSeenRuleSetVersion( 0 ), mRuleSetVersion( 0 )
{
}


ITunesDBSPLRuleSet::ITunesDBSPLRuleSet( const SmartPlaylistRuleSet& src,
                                        ITunesDBPlaylist& playlist,
                                        const ITunesDB& itunesdb )
    : SmartPlaylistRuleSet( src ), mItunesdb( itunesdb ),
      mPlaylist( playlist ),
      mLastSeenItunesDBVersion( 0 ),
      mLastSeenRuleSetVersion( 0 ),
      mRuleSetVersion( 0 )
{
    mRules.clear();
    const RuleList& rules = src.getRules();
    RuleList::const_iterator ruleIter = rules.begin();
    for ( ; ruleIter != rules.end(); ++ruleIter ) {
        if ( (*ruleIter)->isStringBasedRule() ) {
            addStringRule( (SPLStringField)(*ruleIter)->getField(),
                           (SPLStringAction)(*ruleIter)->getAction(),
                           (*ruleIter)->getString() );
        } else {
            addUIntRule( (SPLUIntField)(*ruleIter)->getField(),
                         (SPLUIntAction)(*ruleIter)->getAction(),
                         (*ruleIter)->getFromValue(),
                         (*ruleIter)->getToValue(),
                         (*ruleIter)->getFromDate(),
                         (*ruleIter)->getToDate(),
                         (*ruleIter)->getFromUnits(),
                         (*ruleIter)->getToUnits() );
        }
    }
}


ITunesDBSPLRuleSet::~ITunesDBSPLRuleSet()
{
}


bool ITunesDBSPLRuleSet::operator() ( const itunesdb::Track * track ) const {

    if ( !track || ( !track->isChecked() && checkedTracksOnly() ) ) {
        return false;
    }

    if ( !isCheckRules() ) {
        return true;
    }

    const RuleList& rules = getRules();

    bool noPredicates = true;
    bool result = !mMatchAny;

    RuleList::const_iterator ruleIter = rules.begin();
    for ( ; ruleIter != rules.end() && ( result xor mMatchAny ); ++ruleIter ) {
        TrackPredicate * predicate = dynamic_cast<TrackPredicate*>( *ruleIter );
        if ( predicate ) {
            noPredicates = false;
            if ( mMatchAny ) {
                result |= (*predicate)( track );
            } else {
                result &= (*predicate)( track );
            }
        }
    }

    return result || noPredicates;
}


bool ITunesDBSPLRuleSet::addStringRule( SPLStringField field, SPLStringAction action,
    const QString& testString, int position )
{
    bool success = SmartPlaylistRuleSet::addStringRule( field, action, testString, position );
    if ( success ) {
        mRuleSetVersion++;
    }
    return success;
}


bool ITunesDBSPLRuleSet::addUIntRule( SPLUIntField field, SPLUIntAction action, 
                  Q_UINT64 fromValue, Q_UINT64 toValue,
                   Q_INT64 fromDate,   Q_INT64 toDate,
                  Q_UINT64 fromUnits, Q_UINT64 toUnits,
                  int position )
{
    bool success = SmartPlaylistRuleSet::addUIntRule( field, action, fromValue,
                                toValue, fromDate, toDate, fromUnits, toUnits, position );
    if ( success ) {
        mRuleSetVersion++;
    }
    return success;
}


SmartPlaylistRule *
    ITunesDBSPLRuleSet::createStringRule( SPLStringField field,
                                          SPLStringAction action,
                                          const QString& testString )
{
    const StringPropertyAccessor * stringAccessor = TrackPropertyAccessors::getStringAccessorFor( field );
    if ( stringAccessor ) {
        const StringPredicate * predicate = SPLPredicates::getStringPredicateFor( action );
        if ( predicate ) {
            return new SPLStringTrackPredicate( *predicate, *stringAccessor, testString );
        }
    }

    // unknown - at least it should be written back
    return SmartPlaylistRuleSet::createStringRule( field, action, testString );
}


SmartPlaylistRule *
    ITunesDBSPLRuleSet::createUIntRule( SPLUIntField field, SPLUIntAction action,
                                        Q_UINT64 fromValue, Q_UINT64 toValue,
                                        Q_INT64  fromDate,  Q_INT64  toDate,
                                        Q_UINT64 fromUnits, Q_UINT64 toUnits )
{
    const UIntPropertyAccessor * uintAccessor = TrackPropertyAccessors::getUIntAccessorFor( field );
    if ( uintAccessor ) {
        const UIntPredicate * predicate = SPLPredicates::getUIntPredicateFor( action );
        if ( predicate ) {
            return new SPLUIntTrackPredicate( *predicate, *uintAccessor,
                                              fromValue, toValue,
                                              fromDate, toDate,
                                              fromUnits, toUnits );
        }
    } else if ( field == FIELD_PLAYLISTID ) {
        Playlist * playlist = mItunesdb.getPlaylistByID( fromValue );
        if ( playlist ) {
            return new SPLTrackInPlaylistPredicate( playlist, action,
                                                    fromValue, toValue,
                                                    fromDate, toDate,
                                                    fromUnits, toUnits );
        }
    }

    // unknown - at least it should be written back
    return SmartPlaylistRuleSet::createUIntRule(
                field, action, fromValue, toValue,
                fromDate, toDate, fromUnits, toUnits );
}


void ITunesDBSPLRuleSet::applyLimits( TrackPtrList& trackList ) const {
    if ( !isCheckLimits() ) {
        // do not check limits
        return;
    }

    // calculate limit value
    Q_UINT64 limit = 0;
    switch ( mLimitType ) {
    case 1: limit = 60;   break;
    case 2: limit = 1024; break;
    case 3: limit = 1;    break;
    case 4: limit = 3600; break;
    case 5: limit = 1024 * 1024; break;
    default:
        return;
    }
    limit *= mLimitValue;

    SPLLimitSort limitSort = getLimitSort();

    if ( limitSort == LIMIT_SORT_RANDOM ) {
        trackList.randomize();

    } else {
        TrackComparatorPtr comparator = getComparatorFor( limitSort );
        if ( mReverseLimitSort ) {
            comparator = TrackComparatorPtr( new TrackComparators::ReverseComparator( comparator ) );
        }

        // save the old comparator so we can set it on exit
        const TrackComparatorPtr oldComparator( trackList.getComparator() );
        trackList.setComparator( comparator );

        trackList.sort();

        trackList.setComparator(  oldComparator );
    }

    Q_UINT64 value = 0;
    uint idx = 0;
    while ( idx < trackList.count() ) {
        if ( value >= limit ) {
            trackList.remove( idx );

        } else {
            Track * track = trackList[ idx++ ];
            if ( !track ) {
                continue;
            }
            switch ( mLimitType ) {
            case 1:
            case 4:
                value += track->getTrackLength();
                break;
            case 2:
            case 5:
                value += track->getFileSize(); break;
            case 3:
            default:    // just for completeness
                value++;
                break;
            }
        }
    }
}


TrackComparatorPtr ITunesDBSPLRuleSet::getComparatorFor( SPLLimitSort limitSort ) const {
    switch ( limitSort ) {
    case LIMIT_SORT_RANDOM:
        return TrackComparators::KEEP_ORDER;
    case LIMIT_SORT_SONG_NAME:
        return TrackComparators::BY_TITLE;
    case LIMIT_SORT_ALBUM:
        return TrackComparators::BY_ALBUM;
    case LIMIT_SORT_ARTIST:
        return TrackComparators::BY_ARTIST;
    case LIMIT_SORT_GENRE:
        return TrackComparators::BY_GENRE;
    case LIMIT_SORT_RECENTLYADDED:
        return TrackComparators::BY_RECENTLYADDED;
    case LIMIT_SORT_PLAYCOUNT:
        return TrackComparators::BY_PLAYCOUNT;
    case LIMIT_SORT_LASTPLAYED:
        return TrackComparators::BY_LASTPLAYED_REVERSE;
    case LIMIT_SORT_RATING:
        return TrackComparators::BY_RATING;
    default:
        return TrackComparators::BY_ARTIST;
    }
}


void ITunesDBSPLRuleSet::refreshPlaylist( bool force ) {
    if ( !force ) {
        if ( mLastSeenRuleSetVersion  == mRuleSetVersion &&
             mLastSeenItunesDBVersion == mItunesdb.mInternalVersion )
        {
            return;
        }
    }

    mPlaylist.clear();

    TrackPtrList tracks;
    if ( mItunesdb.getTracksBy( *this, tracks ) ) {
        applyLimits( tracks );
        if ( !tracks.isEmpty() ) {
            mPlaylist.addAll( tracks.iterator() );
        }
    }

    mLastSeenRuleSetVersion = mRuleSetVersion;
    mLastSeenItunesDBVersion = mItunesdb.mInternalVersion;

}

