 /*****************************************************************************
 *   Copyright (C) 2006 by Michael Schulze                                    *
 *   mike.s@genion.de                                                         *
 *                                                                            *
 *  The code contained in this file is free software; you can redistribute    *
 *  it and/or modify it under the terms of the GNU Lesser General Public      *
 *  License as published by the Free Software Foundation; either version      *
 *  2.1 of the License, or (at your option) any later version.                *
 *                                                                            *
 *  This file is distributed in the hope that it will be useful,              *
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of            *
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU         *
 *  Lesser General Public License for more details.                           *
 *                                                                            *
 *  You should have received a copy of the GNU Lesser General Public          *
 *  License along with this code; if not, write to the Free Software          *
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA *
 *                                                                            *
 *  iTunes and iPod are trademarks of Apple                                   *
 *                                                                            *
 *  This product is not supported/written/published by Apple!                 *
 *****************************************************************************/

#include <qdatastream.h>
#include <qbuffer.h>
#include <qdatetime.h>

#include "track.h"
#include "utils.h"

namespace itunesdb {

class TrackComparators::KeepOrder : public TrackComparator {
    virtual int compare( const itunesdb::Track&, const itunesdb::Track& ) const {
        return 0;
    }
};

/**
 * compares two tracks by title and returns a negative value, zero, or a positive value if
 * the first track is lexicographically less, equal or greater than the second.
 */
class TrackComparators::ByTitle : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getTitle().lower().localeAwareCompare( track2.getTitle().lower() );
    }
};

/**
 * compares two tracks by artist and returns a negative value, zero, or a positive value if
 * the first track is lexicographically less, equal or greater than the second.
 */
inline QString getArtistnameForCompare( const itunesdb::Track& track ) {
    QString result( track.getArtistNameForSort() );
    if ( result.isEmpty() ) {
        result = track.getArtist();
    }
    return result.lower();
}
class TrackComparators::ByArtist : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return getArtistnameForCompare( track1 ).localeAwareCompare( getArtistnameForCompare( track2 ) );
    }
};

/**
 * compares two tracks by album and returns a negative value, zero, or a positive value if
 * the first track is lexicographically less, equal or greater than the second.
 */
inline QString getAlbumForCompare( const itunesdb::Track& track ) {
#ifdef USE_ALBUMFORSORT_FOR_SORT
    QString result( track.getAlbumForSort() );
    if ( result.isEmpty() ) {
        result = track.getAlbum();
    }
    return result.lower();
#else
    return track.getAlbum().lower();
#endif
}
class TrackComparators::ByAlbum : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return getAlbumForCompare( track1 ).localeAwareCompare( getAlbumForCompare( track2 ) );
    }
};

/**
 * compares two tracks by genre and returns a negative value, zero, or a positive value if
 * the first track is lexicographically less, equal or greater than the second.
 */
class TrackComparators::ByGenre : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getGenre().lower().localeAwareCompare( track2.getGenre().lower() );
    }
};

/**
 * compares two tracks by category and returns a negative value, zero, or a positive value if
 * the first track is lexicographically less, equal or greater than the second.
 */
class TrackComparators::ByCategory : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getCategory().lower().localeAwareCompare( track2.getCategory().lower() );
    }
};

/**
 * compares two tracks by comment and returns a negative value, zero, or a positive value if
 * the first track is lexicographically less, equal or greater than the second.
 */
class TrackComparators::ByComment : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getComment().localeAwareCompare( track2.getComment() );
    }
};

/**
 * compares two tracks by composer and returns a negative value, zero, or a positive value if
 * the first track is lexicographically less, equal or greater than the second.
 */
class TrackComparators::ByComposer : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getComposer().lower().localeAwareCompare( track2.getComposer().lower() );
    }
};

/**
 * compares two tracks by equalizer setting and returns a negative value, zero, or a positive value if
 * the first track is lexicographically less, equal or greater than the second. This basically groups songs
 * with the same setting together
 */
class TrackComparators::ByEqualizerSetting : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getEqualizerSetting().localeAwareCompare( track2.getEqualizerSetting() );
    }
};

/**
 * compares two tracks by description and returns a negative value, zero, or a positive value if
 * the first track is lexicographically less, equal or greater than the second.
 */
class TrackComparators::ByDescriptionTxt : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getDescription().localeAwareCompare( track2.getDescription() );
    }
};

/**
 * compares two tracks by grouping and returns a negative value, zero, or a positive value if
 * the first track is lexicographically less, equal or greater than the second.
 */
class TrackComparators::ByGrouping : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getGrouping().localeAwareCompare( track2.getGrouping() );
    }
};

/**
 * compares two tracks by bitrate and returns a negative value, zero, or a positive value if
 * the first track is less, equal or greater than the second.
 */
class TrackComparators::ByBitrate : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getBitrate() - track2.getBitrate();
    }
};

/**
 * compares two tracks by file type and returns a negative value, zero, or a positive value if
 * the first track is less, equal or greater than the second. This basically groups songs
 * with the same filetype together
 */
class TrackComparators::ByFiletype : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getFileFormatCode() - track2.getFileFormatCode();
    }
};

/**
 * compares two tracks by modification time and returns a negative value, zero, or a positive value if
 * the first track is less, equal or greater than the second.
 */
class TrackComparators::ByTimeModified : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getLastModified() - track2.getLastModified();
    }
};

/**
 * compares two tracks by track number and returns a negative value, zero, or a positive value if
 * the first track is less, equal or greater than the second.
 */
class TrackComparators::ByTrackNumber : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getTrackNumber() - track2.getTrackNumber();
    }
};

/**
 * compares two tracks by size and returns a negative value, zero, or a positive value if
 * the first track is less, equal or greater than the second.
 */
class TrackComparators::BySize : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getFileSize() - track2.getFileSize();
    }
};

/**
 * compares two tracks by length and returns a negative value, zero, or a positive value if
 * the first track is less, equal or greater than the second.
 */
class TrackComparators::ByTrackLength : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getTrackLength() - track2.getTrackLength();
    }
};

/**
 * compares two tracks by year and returns a negative value, zero, or a positive value if
 * the first track is less, equal or greater than the second.
 */
class TrackComparators::ByYear : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getYear() - track2.getYear();
    }
};

/**
 * compares two tracks by sample rate and returns a negative value, zero, or a positive value if
 * the first track is less, equal or greater than the second.
 */
class TrackComparators::BySampleRate : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getSamplerate() - track2.getSamplerate();
    }
};

/**
 * compares two tracks by the date added to the ipod and returns a negative value, zero, or
 * a positive value if the first track is less, equal or greater than the second.
 */
class TrackComparators::ByDateAdded : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getDateAdded() - track2.getDateAdded();
    }
};

/**
 * compares two tracks by playcount and returns a negative value, zero, or a positive value if
 * the first track is less, equal or greater than the second.
 */
class TrackComparators::ByPlaycount : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track2.getPlayCount() - track1.getPlayCount();
    }
};

/**
 * compares two tracks by last played time and returns a negative value, zero, or a positive value if
 * the first track is less, equal or greater than the second.
 */
class TrackComparators::ByLastPlayed : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getLastPlayed() - track2.getLastPlayed();
    }
};

/**
 * compares two tracks by rating and returns a negative value, zero, or a positive value if
 * the first track is less, equal or greater than the second.
 */
class TrackComparators::ByRating : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getRating() - track2.getRating();
    }
};

/**
 * compares two tracks by cd number and returns a negative value, zero, or a positive value if
 * the first track is less, equal or greater than the second.
 */
class TrackComparators::ByCdNumber : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getCdNumber() - track2.getCdNumber();
    }
};

/**
 * compares two tracks by release date and returns a negative value, zero, or a positive value if
 * the first track is less, equal or greater than the second.
 */
class TrackComparators::ByReleaseDate : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getReleaseDate() - track2.getReleaseDate();
    }
};

/**
 * compares two tracks by beats per minute info and returns a negative value,
 * zero, or a positive value if the first track is less, equal or greater than
 * the second.
 */
class TrackComparators::ByBPM : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getBPM() - track2.getBPM();
    }
};

/**
 * compares two tracks by skipcount and returns a negative value, zero, or a positive value if
 * the first track is less, equal or greater than the second.
 */
class TrackComparators::BySkipCountOnly : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track2.getSkipCount() - track1.getSkipCount();
    }
};

/**
 * compares two tracks by last skip time and returns a negative value, zero, or a positive value if
 * the first track is less, equal or greater than the second.
 */
class TrackComparators::ByLastSkipped : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getLastSkippedTime() - track2.getLastSkippedTime();
    }
};

/**
 * compares two tracks by Tv Show name and returns a negative value, zero, or a positive value if
 * the first track is less, equal or greater than the second.
 */
class TrackComparators::ByTvShow : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getTvShow().localeAwareCompare( track2.getTvShow() );
    }
};

/**
 * compares two tracks by episode number and returns a negative value, zero, or a positive value if
 * the first track is less, equal or greater than the second.
 */
class TrackComparators::BySeason : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getSeasonNumber() - track2.getSeasonNumber();
    }
};

/**
 * compares two tracks by episode number and returns a negative value, zero, or a positive value if
 * the first track is less, equal or greater than the second.
 */
class TrackComparators::ByEpisodeNumber : public TrackComparator {
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        return track1.getEpisodeNumber() - track2.getEpisodeNumber();
    }
};

TrackComparatorPtr TrackComparators::KEEP_ORDER =
    TrackComparatorPtr( new TrackComparators::KeepOrder() );
TrackComparatorPtr TrackComparators::BY_TITLE =
    TrackComparatorPtr( new TrackComparators::ByTitle() );
TrackComparatorPtr TrackComparators::BY_ARTISTONLY =
    TrackComparatorPtr( new TrackComparators::ByArtist() );
TrackComparatorPtr TrackComparators::BY_ALBUMONLY =
    TrackComparatorPtr( new TrackComparators::ByAlbum() );
TrackComparatorPtr TrackComparators::BY_TVSHOWONLY =
    TrackComparatorPtr( new TrackComparators::ByTvShow() );
TrackComparatorPtr TrackComparators::BY_SEASONONLY =
    TrackComparatorPtr( new TrackComparators::BySeason() );
TrackComparatorPtr TrackComparators::BY_EPISODEONLY =
    TrackComparatorPtr( new TrackComparators::ByEpisodeNumber() );
TrackComparatorPtr TrackComparators::BY_CATEGORY =
    TrackComparatorPtr(
        ( new TrackComparators::ComposingComparator( new TrackComparators::ByCategory() ) )
            ->add( BY_TRACKNUMBER )
            ->add( BY_TITLE )
    );
TrackComparatorPtr TrackComparators::BY_FILETYPE =
    TrackComparatorPtr( new TrackComparators::ByFiletype() );
TrackComparatorPtr TrackComparators::BY_TRACKLENGTH =
    TrackComparatorPtr( new TrackComparators::ByTrackLength() );
TrackComparatorPtr TrackComparators::BY_YEAR =
    TrackComparatorPtr( new TrackComparators::ByYear() );
TrackComparatorPtr TrackComparators::BY_COMMENT =
    TrackComparatorPtr( new TrackComparators::ByComment() );
TrackComparatorPtr TrackComparators::BY_COMPOSERONLY =
    TrackComparatorPtr( new TrackComparators::ByComposer() );
TrackComparatorPtr TrackComparators::BY_TRACKNUMBER =
    TrackComparatorPtr( new TrackComparators::ByTrackNumber() );
TrackComparatorPtr TrackComparators::BY_SIZE =
    TrackComparatorPtr( new TrackComparators::BySize() );
TrackComparatorPtr TrackComparators::BY_TIMEMODIFIED =
    TrackComparatorPtr( new TrackComparators::ByTimeModified() );
TrackComparatorPtr TrackComparators::BY_GENREONLY =
    TrackComparatorPtr( new TrackComparators::ByGenre() );
TrackComparatorPtr TrackComparators::BY_BITRATE =
    TrackComparatorPtr( new TrackComparators::ByBitrate() );
TrackComparatorPtr TrackComparators::BY_SAMPLERATE =
    TrackComparatorPtr( new TrackComparators::BySampleRate() );
TrackComparatorPtr TrackComparators::BY_DATEADDED =
    TrackComparatorPtr( new TrackComparators::ByDateAdded() );
TrackComparatorPtr TrackComparators::BY_RECENTLYADDED =
    TrackComparatorPtr( new TrackComparators::ReverseComparator( BY_DATEADDED ) );
TrackComparatorPtr TrackComparators::BY_DESCRIPTIONTXT =
    TrackComparatorPtr(
        ( new TrackComparators::ComposingComparator( new TrackComparators::ByDescriptionTxt() ) )
            ->add( BY_TRACKNUMBER )
            ->add( BY_TITLE )
    );
TrackComparatorPtr TrackComparators::BY_GROUPING =
    TrackComparatorPtr( new TrackComparators::ByGrouping() );
TrackComparatorPtr TrackComparators::BY_EQSETTING = 
    TrackComparatorPtr( new TrackComparators::ByEqualizerSetting() );
TrackComparatorPtr TrackComparators::BY_PLAYCOUNTONLY =
    TrackComparatorPtr( new TrackComparators::ByPlaycount() );
TrackComparatorPtr TrackComparators::BY_LASTPLAYED =
    TrackComparatorPtr( new TrackComparators::ByLastPlayed() );
TrackComparatorPtr TrackComparators::BY_LASTPLAYED_REVERSE =
    TrackComparatorPtr( new TrackComparators::ReverseComparator( BY_LASTPLAYED ) );
TrackComparatorPtr TrackComparators::BY_CDNUMBER =
    TrackComparatorPtr( new TrackComparators::ByCdNumber() );
TrackComparatorPtr TrackComparators::BY_RATING =
    TrackComparatorPtr( new TrackComparators::ByRating() );
TrackComparatorPtr TrackComparators::BY_RELEASEDATE =
    TrackComparatorPtr( new TrackComparators::ByReleaseDate() );
TrackComparatorPtr TrackComparators::BY_BPM =
    TrackComparatorPtr( new TrackComparators::ByBPM() );
TrackComparatorPtr TrackComparators::BY_SKIPCOUNTONLY =
    TrackComparatorPtr( new TrackComparators::BySkipCountOnly() );
TrackComparatorPtr TrackComparators::BY_LASTSKIPPED =
    TrackComparatorPtr( new TrackComparators::ByLastSkipped() );
TrackComparatorPtr TrackComparators::BY_LASTSKIPPED_REVERSE =
    TrackComparatorPtr( new TrackComparators::ReverseComparator( BY_LASTSKIPPED ) );

TrackComparatorPtr TrackComparators::BY_ALBUM = TrackComparatorPtr(
    ( new TrackComparators::ComposingComparator( BY_ALBUMONLY ) )
        ->add( BY_CDNUMBER )
        ->add( BY_TRACKNUMBER )
        ->add( BY_TITLE ) );

TrackComparatorPtr TrackComparators::BY_ARTIST = TrackComparatorPtr(
    ( new TrackComparators::ComposingComparator( BY_ARTISTONLY ) )
        ->add( BY_ALBUM ) );

TrackComparatorPtr TrackComparators::BY_GENRE = TrackComparatorPtr(
    ( new TrackComparators::ComposingComparator( BY_GENREONLY ) )
        ->add( BY_ARTIST ) );

TrackComparatorPtr TrackComparators::BY_COMPOSER = TrackComparatorPtr(
    ( new TrackComparators::ComposingComparator( BY_COMPOSERONLY ) )
        ->add( BY_TITLE ) );

TrackComparatorPtr TrackComparators::BY_PLAYCOUNT = TrackComparatorPtr(
    ( new TrackComparators::ComposingComparator( BY_PLAYCOUNTONLY ) )
        ->add( BY_ARTIST ) );

TrackComparatorPtr TrackComparators::BY_SKIPCOUNT = TrackComparatorPtr(
    ( new TrackComparators::ComposingComparator( BY_SKIPCOUNTONLY ) )
        ->add( BY_ARTIST ) );

TrackComparatorPtr TrackComparators::BY_EPISODE = TrackComparatorPtr(
    ( new TrackComparators::ComposingComparator( BY_EPISODEONLY ) )
        ->add( BY_TITLE ) );

TrackComparatorPtr TrackComparators::BY_SEASON = TrackComparatorPtr(
    ( new TrackComparators::ComposingComparator( BY_SEASONONLY ) )
        ->add( BY_EPISODE ) );

TrackComparatorPtr TrackComparators::BY_TVSHOW = TrackComparatorPtr(
    ( new TrackComparators::ComposingComparator( BY_TVSHOWONLY ) )
        ->add( BY_SEASON ) );

TrackPredicates::AllTracks TrackPredicates::ALLTRACKS = TrackPredicates::AllTracks();

Track::Track( Q_UINT32 trackid ) :
    ListItem( ITEMTYPE_TRACK ),
    id( trackid ),
    lastmodified( 0 ),
    file_size( 0 ),
    tracklen( 0 ),
    tracknum( 0 ),
    numtracks( 0 ),
    year( 0 ),
    bitrate( 0 ),
    samplerate( 0xAC440000 ), // hardwired for now
    volumeadjust( 0 ),
    playcount( 0 ),
    last_played_at( 0 ),
    cdnum( 0 ),
    numcds( 0 ),
    file_format_code( 0x0C ),
    date_added( 0 ),
    date_released( 0 ),
    rating ( 0 ),
    vbr( 0 ),
    type( 0 ),
    compilation( 0 ),
    dbid( 0 ),
    checked( 1 ),
    app_rating( 0 ),
    beatsPerMinute( 0 ),
    mArtworkCount( 0 ),
    unk9( 0xFFFF ),
    mArtworkSize( 0 ),
    unk11( 0 ),
    mAppleDRM( 0 ),
    mSkipCount( 0 ),
    mLastSkippedDate( 0 ),
    mHasArtwork( 0x02 ),    // no artwork by default
    mSkipWhenShuffe( 0 ),
    mRememberPbackPos( 0 ),
    flag4( 0 ),
    mHasLyricsFlag( 0 ),
    mIsMovie( 0 ),
    mPodcastPlayedFlag( 1 ),
    unk37_2( 0 ),
    unk21( 0 ),
    mPregapSamples( 0 ),
    mNumSamples( 0 ),
    unk25( 0 ),
    mPostgapSamples( 0 ),
    unk27( 0 ),
    mMediaType( Track::Audio ),
    mSeasonNumber( 0 ),
    mEpisodeNumber( 0 ),
    unk31( 0 ), // seems to be always 0 for unprotected files
    unk32( 0 ),
    unk33( 0 ),
    unk34( 0 ),
    unk35( 0 ),
    unk36( 0 ),
    unk37( 0 ),
    mGaplessData( 0 ),
    unk38( 0 ),
    mGapLessFlag( 0 ),
    mGaplessAlbum( 0 ), // no crossfading in iTunes
    m_recentlyPlayed( false )
{
    setFileFormatDesc( QString( "MPEG audio file" ) );    // hardwired for now
}

// 
Track::~Track() {
}


void Track::writeData( QByteArray& data ) const {
    // FIXME make IO symmetric
#if (QT_VERSION < 0x040000)
    QBuffer buffer( data );
#else
    QBuffer buffer( &data );
#endif
    buffer.open(IO_WriteOnly);
    QDataStream stream( &buffer);
    stream.setByteOrder( QDataStream::LittleEndian);

    /** Write the track header **/
    stream << (Q_UINT32) 0x7469686D;        // 0x00 mhit
    stream << (Q_UINT32) 0x104;              // 0x04 headerlen
    stream << (Q_UINT32) 0x0;               // 0x08 length - set later
    stream << (Q_UINT32) 0x0;               // 0x0C number of mhods
    stream << (Q_UINT32) getID();           // 0x10
    stream << (Q_UINT32) 1;                 // 0x14
    stream << (Q_UINT32) 0x4d503320;        // 0x18 ipod shuffle wants a "MP3 " here
    stream << vbr;                          // 0x1C
    stream << type;                         // 0x1D
    stream << compilation;                  // 0x1E
    stream << rating;                       // 0x1F
    stream << (Q_UINT32) (getLastModified() ? getLastModified() + MAC_EPOCH_DELTA : 0); // 0x20
    stream << (Q_UINT32) getFileSize();     // 0x24
    stream << (Q_UINT32) getTrackLength();  // 0x28
    stream << (Q_UINT32) getTrackNumber();  // 0x2C
    stream << (Q_UINT32) getTrackCount();   // 0x30
    stream << (Q_UINT32) getYear();         // 0x34
    stream << (Q_UINT32) getBitrate();      // 0x38
    stream << (Q_UINT32) getSamplerate();   // 0x3C
    stream << (Q_UINT32) getVolumeAdjust(); // 0x40
    stream << (Q_UINT32) getStartTime();    // 0x44 start time
    stream << (Q_UINT32) getStopTime();     // 0x48
    stream << (Q_UINT32) getSoundCheck();   // 0x4C
    stream << (Q_UINT32) getPlayCount();    // 0x50
    stream << (Q_UINT32) getPlayCount();    // 0x54
    stream << (Q_UINT32) (getLastPlayed() ? getLastPlayed() + MAC_EPOCH_DELTA : 0);   // 0x58
    stream << (Q_UINT32) getCdNumber();     // 0x5C
    stream << (Q_UINT32) getCdCount();      // 0x60
    stream << (Q_UINT32) 0;                 // 0x64 empty space //userid from apple store
    stream << (Q_UINT32) ( date_added ? date_added + MAC_EPOCH_DELTA : 0 ); // 0x68
    stream << (Q_UINT32) 0;                 // bookmarktime
    stream << (Q_UINT64) dbid;              // unique bit (64 bit)
    stream << (Q_UINT8)  checked;           // checked in iTunes
    stream << (Q_UINT8)  app_rating;        // application rating
    stream << (Q_UINT16) beatsPerMinute;    // BPM
    stream << (Q_UINT16) mArtworkCount;     // artworkcount
    stream << (Q_UINT16) unk9;              // unkown
    stream << (Q_UINT32) mArtworkSize;      // artwork size
    stream << (Q_UINT32) unk11;             // unkown
    stream << (float) -getSamplerate();     // 0x88 - samplerate as floating point "-"?!?
    stream << (Q_UINT32) date_released;     // 0x8C - release date/time
    stream << (Q_UINT32) file_format_code;  // 0x90 - unkown, but 0x0000000c for MP3 ?
    stream << (Q_UINT32) mAppleDRM;         // 0x94 - used for Apple Store DRM songs, 0 otherwise
    stream << (Q_UINT32) 0;                 // 0x98 - unkown
    stream << (Q_UINT32) mSkipCount;        // 0x9A - unkown
    stream << (Q_UINT32) ( mLastSkippedDate ? mLastSkippedDate + MAC_EPOCH_DELTA : 0 );  // 0xA0 - unkown
    stream << (Q_UINT8)  mHasArtwork;       // 0xA4 0x01 if track has artwork, 0x02 otherwise
    stream << (Q_UINT8)  mSkipWhenShuffe;   // 0xA5 0x01 if track should be skipped when shuffling
    stream << (Q_UINT8)  mRememberPbackPos; // 0xA6 0x01 if remember playback position is set
    stream << (Q_UINT8)  flag4;              // set to 1 or 2 if podcast, 0 otherwise
    stream << (Q_UINT64) dbid;              // 0xA8 - same unique id as above
    stream << (Q_UINT8)  mHasLyricsFlag;    // 0xB0 - 0x01 if lyrics stored in MP3 meta tags
    stream << (Q_UINT8)  mIsMovie;          // 0xB1 - 0x01 if refers to a movie, 0x0 otherwise
    stream << (Q_UINT8)  mPodcastPlayedFlag;// 0xB2 - 0x1 if not a podcast, 0x2 if podcast not played yet
    stream << (Q_UINT8)  unk37_2;           // 0xB3 - so far always 0
    stream << (Q_UINT32) unk21;             // 0xB4 unk21
    stream << (Q_UINT32) mPregapSamples;    // 0xB8 num of pregap samples - gapless playback supp
    stream << (Q_UINT64) mNumSamples;       // 0xBC num of samples in song - gapless playback supp
    stream << (Q_UINT32) unk25;             // 0xC4 unk25
    stream << (Q_UINT32) mPostgapSamples;   // 0xC8 num of postgap samples - gapless playack supp
    stream << (Q_UINT32) unk27;             // 0xCC
    stream << (Q_UINT32) mMediaType;        // 0xD0 MediaType - 0x01 for audio files
    stream << (Q_UINT32) mSeasonNumber;     // 0xD4 Season number for TV shows
    stream << (Q_UINT32) mEpisodeNumber;    // 0xD8 Episode number for TV shows
    stream << (Q_UINT32) unk31;             // 0xDC seems to be always 0 for unprotected files
    stream << (Q_UINT32) unk32;             // 0xE0
    stream << (Q_UINT32) unk33;             // 0xE4
    stream << (Q_UINT32) unk34;             // 0xE8
    stream << (Q_UINT32) unk35;             // 0xEC
    stream << (Q_UINT32) unk36;             // 0xF0
    stream << (Q_UINT32) unk37;             // 0xF4
    stream << (Q_UINT32) mGaplessData;      // 0xF8 size from first Synch Frame until the 8th before the last frame
    stream << (Q_UINT32) unk38;             // 0xFC
    stream << (Q_UINT16) mGapLessFlag;      // 0x100 set to 1 if gapless playback is desired
    stream << (Q_UINT16) mGaplessAlbum;     // 0x102 set to 1 if album is gapless

    /** Write Track contents **/
    Q_UINT32 num_mhods = writeMHODsToStream( stream );

    buffer.at( 8);
    stream << (Q_UINT32)data.size();	// set content length
    stream << (Q_UINT32)num_mhods;	// set real mhod count
    buffer.close();
}

QDataStream & Track::writeToStream(QDataStream & outstream ) {
    QByteArray buffer;
    writeData( buffer );
    outstream.writeRawBytes( buffer.data(), buffer.size() );

    return outstream;
}


inline void checkMacEpochDate( Q_UINT32& date ) {
    if ( date ) {
        date -= MAC_EPOCH_DELTA;
    }
}

QDataStream & Track::readFromStream( QDataStream& instream, bool * ok ) {
    Q_UINT32 dummy, blocklen;
    instream >> blocklen;

    if ( blocklen < 108 ) {
        // insufficient information
        QByteArray buffer( blocklen );
        instream.readRawBytes( buffer.data(), blocklen - 8 );
        if ( ok ) {
            *ok = false;
        }
        return instream;
    }

    instream >> dummy;
    instream >> dummy;  // number of mhods doesn't really sound interesting
    instream >> id;     // 16

    instream >> dummy;  // 20
    instream >> dummy;  // 24
    instream >> vbr;    // 28
    instream >> type;   // 29
    instream >> compilation;    // 30
    instream >> rating; // 31
    instream >> lastmodified;   // 32
    checkMacEpochDate( lastmodified );

    instream >> file_size;

    instream >> tracklen;
    instream >> tracknum;
    instream >> numtracks;
    instream >> year;
    instream >> bitrate;

    instream >> samplerate;
    instream >> volumeadjust;
    instream >> mStartTime;
    instream >> mStopTime;
    instream >> mSoundCheck;

    instream >> playcount;
    instream >> dummy;
    instream >> last_played_at;
    checkMacEpochDate( last_played_at );

    instream >> cdnum;  // 94
    instream >> numcds; // 96

    instream >> dummy;  // 100
    instream >> date_added; // 104
    checkMacEpochDate( date_added );

    // 108 byte read so far

    if ( blocklen >= 148 ) { // iTunes >= 4.7
        instream >> dummy;  // 108
        instream >> dbid;   // 112
        instream >> checked;    // 120;
        instream >> app_rating; // 121;
        instream >> beatsPerMinute; // 122
        instream >> mArtworkCount;  // 124
        instream >> unk9;           // 126
        instream >> mArtworkSize;   // 128
        instream >> dummy;          // 132
        instream >> dummy;          // 136  // samplerate
        instream >> date_released;  // 140
        // FIXME is the mac epoch date handled elsewhere?
        instream >> file_format_code;   // 144

        // 148 byte read so far
        if ( blocklen >= 260 ) {
            instream >> mAppleDRM;      // 148
            instream >> dummy;          // 152
            instream >> mSkipCount;     // 156
            instream >> mLastSkippedDate;   // 160
            checkMacEpochDate( mLastSkippedDate );
            instream >> mHasArtwork;    // 164
            instream >> mSkipWhenShuffe;    // 165
            instream >> mRememberPbackPos;  // 166
            instream >> flag4;          // 167
            instream >> dummy;          // 168 dbid first quad
            instream >> dummy;          // 172 dbid second quad
            instream >> mHasLyricsFlag; // 176
            instream >> mIsMovie;       // 177
            instream >> mPodcastPlayedFlag; // 178
            instream >> unk37_2;        // 179
            instream >> unk21;          // 180
            instream >> mPregapSamples; // 184
            instream >> mNumSamples;    // 188
            instream >> unk25;          // 196
            instream >> mPostgapSamples;// 200
            instream >> unk27;          // 204
            instream >> mMediaType;     // 208
            instream >> mSeasonNumber;  // 212
            instream >> mEpisodeNumber; // 216
            instream >> unk31;          // 220
            instream >> unk32;          // 224
            instream >> unk33;          // 228
            instream >> unk34;          // 232
            instream >> unk35;          // 236
            instream >> unk36;          // 240
            instream >> unk37;          // 244
            instream >> mGaplessData;   // 248
            instream >> unk38;          // 252
            instream >> mGapLessFlag;   // 256
            instream >> mGaplessAlbum;  // 258

            // 260 byte read so far - ignore the rest for now
            itunesdb::utils::seekRelative( instream, blocklen - 260 );

        } else {
            // 148 byte read so far - ignore the rest for now
            itunesdb::utils::seekRelative( instream, blocklen - 148 );

        }
    } else {
        itunesdb::utils::seekRelative( instream, blocklen - 108 );
    }

    setDirty( false );
    if ( ok ) {
        *ok = true;
    }
    return instream;
}


Q_UINT32 Track::getID() const {
    return id;
}

const QString & Track::getAlbum() const
{
    return getItemProperty( MHOD_ALBUM);
}


const QString& Track::getAlbumForSort() const {
    return getItemProperty( itunesdb::MHOD_ALBUM_FOR_SORT );
}


const QString & Track::getArtist() const
{
    return getItemProperty( MHOD_ARTIST);
}


QString Track::getPlainArtistName() const
{
    QString result = getItemProperty( MHOD_ARTIST_FOR_SORT );
    if ( result.isEmpty() ) {
        result = getArtist();
    }

    return result;
}


Q_UINT16 Track::getBPM() const {
    return beatsPerMinute;
}


const QString& Track::getCategory() const {
    return getItemProperty( MHOD_CATEGORY );
}


const QString & Track::getComment() const
{
    return getItemProperty( MHOD_COMMENT);
}


const QString & Track::getTvShow() const
{
    return getItemProperty( MHOD_TVSHOW );
}


void Track::setTvShow( const QString& show )
{
    setItemProperty( show, MHOD_TVSHOW );
}


const QString & Track::getTitle() const
{
    return getItemProperty( MHOD_TITLE);
}

const QString & Track::getTitleForSort() const
{
    return getItemProperty( itunesdb::MHOD_TITLE_FOR_SORT );
}


const QString & Track::getGenre() const
{
    return getItemProperty( MHOD_GENRE);
}


const QString & Track::getPathInfo() const
{
    return getItemProperty( itunesdb::MHOD_PATH);
}

const QString & Track::getComposer() const
{
    return getItemProperty( itunesdb::MHOD_COMPOSER);
}

const QString & Track::getEqualizerSetting() const {
    return getItemProperty( itunesdb::MHOD_EQ_SETTING );
}

const QString& Track::getDescription() const {
    return getItemProperty( itunesdb::MHOD_DESCRIPTION_TEXT );
}

const QString& Track::getGrouping() const {
    return getItemProperty( itunesdb::MHOD_GROUPING );
}

const QString& Track::getAlbumArtist() const {
    return getItemProperty( itunesdb::MHOD_ALBUMARTIST );
}

const QString& Track::getAlbumArtistForSort() const {
    return getItemProperty( itunesdb::MHOD_ALBUMARTIST_FOR_SORT );
}

const QString& Track::getArtistNameForSort() const {
    return getItemProperty( itunesdb::MHOD_ARTIST_FOR_SORT );
}

const QString& Track::getKeywords() const {
    return getItemProperty( itunesdb::MHOD_KEYWORDS );
}

const QString& Track::getFileFormatDesc() const {
    return getItemProperty( itunesdb::MHOD_FDESC );
}

void Track::setRating(unsigned char rating) {
    this->rating = rating;
}

unsigned char Track::getRating() const {
    return rating;
}

Q_UINT32 Track::getLastModified() const {
    return lastmodified;
}

Q_UINT32 Track::getFileSize() const {
    return file_size;
}

Q_UINT32 Track::getTrackLength() const {
    return tracklen;
}

Q_UINT32 Track::getTrackNumber() const {
    return tracknum;
}

Q_UINT32 Track::getTrackCount() const {
    return numtracks;
}

Q_UINT32 Track::getYear() const {
    return year;
}

Q_UINT32 Track::getBitrate() const {
    return bitrate;
}

Q_UINT32 Track::getSamplerate() const {
    return samplerate;
}

Q_UINT32 Track::getVolumeAdjust() const {
    return volumeadjust;
}

void Track::setPlayCount(Q_UINT32 newPlaycount) {
    this->playcount = newPlaycount;
}

Q_UINT32 Track::getPlayCount() const {
    return playcount;
}

void Track::setLastPlayed(Q_UINT32 lastplayed) {
    this->last_played_at = lastplayed;
}

Q_UINT32 Track::getLastPlayed() const {
    return last_played_at;
}

Q_UINT32 Track::getCdNumber() const {
    return cdnum;
}

Q_UINT32 Track::getCdCount() const {
    return numcds;
}

unsigned char Track::getVbrFlag() const {
    return vbr;
}

bool Track::isVideo() const {
    switch( mMediaType ) {
    case Audio_Video:
    case Video:
    case Video_Podcast:
    case Music_Video:
    case TV_Show:
    case TV_Show_w_Audio:
        return true;
    default:
        return false;
    }
}

/**
 */
Q_UINT64 Track::getDBID() const
{
    return dbid;
}


void Track::setAlbum( const QString& album ) {
    setItemProperty( album, itunesdb::MHOD_ALBUM);
}


void Track::setAlbumForSort( const QString& album ) {
    setItemProperty( album, itunesdb::MHOD_ALBUM_FOR_SORT );
}


inline uint calculatePlainArtistnameStartIdx( const QString& artist ) {
    QString first4 = artist.left( 4 ).lower();
    if ( first4.compare( "the " ) == 0 ||
         first4.compare( "der " ) == 0 ||
         first4.compare( "die " ) == 0 ||
         first4.compare( "das " ) == 0 )
    {
        return 4;
    }
    // TODO question to our french and spanish speaking friends
    // TODO would it be right to care about "le" and "el" here?
    return 0;
}

void Track::setArtist( const QString& artist ) {
    setItemProperty( artist, itunesdb::MHOD_ARTIST);
    uint plainArtistNameStartIdx = calculatePlainArtistnameStartIdx( artist );
    if ( plainArtistNameStartIdx != 0 ) {
        QString name = artist.mid( plainArtistNameStartIdx );
        QString article = artist.left( plainArtistNameStartIdx ).stripWhiteSpace();
        setArtistNameForSort( name );
    } else {
        setArtistNameForSort( QString::null );
    }
}

void Track::setArtistNameForSort( const QString& artist ) {
    setItemProperty( artist, itunesdb::MHOD_ARTIST_FOR_SORT );
}

void Track::setCategory( const QString& category ) {
    setItemProperty( category, itunesdb::MHOD_CATEGORY);
}


void Track::setGenre( const QString& genre ) {
    setItemProperty( genre, itunesdb::MHOD_GENRE);
}


void Track::setTitle( const QString& title ) {
    setItemProperty( title, itunesdb::MHOD_TITLE);
}


void Track::setTitleForSort( const QString& title ) {
    setItemProperty( title, itunesdb::MHOD_TITLE_FOR_SORT );
}


void Track::setFileFormatDesc( const QString& fdesc ) {
    setItemProperty( fdesc, itunesdb::MHOD_FDESC);
}

void Track::setComposer( const QString& composer ) {
    setItemProperty( composer, itunesdb::MHOD_COMPOSER);
}

void Track::setDescription( const QString& description_txt ) {
    setItemProperty( description_txt, itunesdb::MHOD_DESCRIPTION_TEXT );
}

void Track::setGrouping( const QString& grouping ) {
    setItemProperty( grouping, itunesdb::MHOD_GROUPING );
}

void Track::setEqualizerSetting( const QString& eqSetting ) {
    setItemProperty( eqSetting, itunesdb::MHOD_EQ_SETTING );
}

void Track::setAlbumArtist( const QString& albumartist ) {
    setItemProperty( albumartist, itunesdb::MHOD_ALBUMARTIST );
}

void Track::setAlbumArtistForSort( const QString& albumartist ) {
    setItemProperty( albumartist, itunesdb::MHOD_ALBUMARTIST_FOR_SORT );
}

void Track::setKeywords( const QString& keywords ) {
    setItemProperty( keywords, itunesdb::MHOD_KEYWORDS );
}

void Track::setCompilationFlag(unsigned char isCompilation) {
    compilation = isCompilation;
}

void Track::setTrackNumber(Q_UINT32 tracknumber) {
    tracknum = tracknumber;
    setDirty();
}

void itunesdb::Track::setTrackCount(Q_UINT32 trackcount) {
    numtracks = trackcount;
    // don't set to dirty this is only updated when tracks are added
}


/**
 */
void Track::setDBID( Q_UINT64 id )
{
    dbid = id;
}


void Track::setComment( const QString& comment ) {
    setItemProperty( comment, itunesdb::MHOD_COMMENT );
}


void Track::setPathInfo( const QString& encodedpath ) {
    setItemProperty( encodedpath, itunesdb::MHOD_PATH );
}

void Track::setVolumeAdjust( const Q_UINT32 newVolumeAdjust ) {
    volumeadjust = newVolumeAdjust;
    setDirty();
}

void Track::setYear( const Q_UINT32 newYear ) {
    year = newYear;
    setDirty();
}

void Track::setStartTime( Q_UINT32 startTime ) {
    mStartTime = startTime;
    setDirty();
}

void Track::setStopTime( Q_UINT32 stopTime ) {
    mStopTime = stopTime;
    setDirty();
}

void Track::setSoundCheck( Q_UINT32 soundCheck ) {
    mSoundCheck = soundCheck;
    setDirty();
}

QString Track::getFileExtension() const {
    QString filename = getPathInfo();
    if ( !filename.isEmpty() ) {
        return filename.section('.', -1, -1, QString::SectionIncludeLeadingSep);
    }
    return filename;
}


void Track::setFileExtension( const QString& fileextension ) {
    QString filename = getPathInfo();
    if ( !filename.isEmpty() ) {
        int start = filename.findRev( '.' );
        if ( start == -1 ) {
            if ( !fileextension.startsWith( "." ) ) {
                filename.append( '.' );
            }
            filename.append( fileextension );
        } else {
            if ( !fileextension.startsWith( "." ) ) {
                start++;
            }
            filename.replace( start, filename.length() - start, fileextension );
        }
    }
    setPathInfo( filename );
}


void Track::setSkipCount( Q_UINT32 numberOfSkips ) {
    mSkipCount = numberOfSkips;
}


Q_UINT32 Track::getSkipCount() const {
    return mSkipCount;
}


void Track::setLastSkippedTime( Q_UINT32 lastSkippedDateTime ) {
    mLastSkippedDate = lastSkippedDateTime;
}

Q_UINT32 Track::getLastSkippedTime() const {
    return mLastSkippedDate;
}


void Track::skipWhenShuffle( bool enable ) {
    mSkipWhenShuffe = enable;
}


bool Track::skippedWhenShuffle() const {
    return mSkipWhenShuffe;
}


void Track::setHasLyrics( bool hasLyrics ) {
    mHasLyricsFlag = hasLyrics ? 1 : 0;
}


bool Track::hasLyrics() const {
    return mHasLyricsFlag;
}


void Track::setMediaType( Track::MediaType type ) {
    mMediaType = type;
    if ( type == Video || type == Audio_Video || type == Video_Podcast ) {
        mIsMovie = true;
    } else {
        mIsMovie = false;
    }
}


Track::MediaType Track::getMediaType() const {
    return ( Track::MediaType ) mMediaType;
}


void Track::setGaplessPBackData( Q_UINT32 pre, Q_UINT64 song, Q_UINT32 post, Q_UINT32 data ) {
    if ( mPregapSamples == 0 && mNumSamples == 0 && mPostgapSamples == 0 ) {
        mGapLessFlag = 0;
    } else {
        mGapLessFlag = 1;
    }

    mPregapSamples = pre;
    mNumSamples = song;
    mPostgapSamples = post;
    mGaplessData = data;
}


Q_UINT32 Track::getNumPregapSamples() const {
    return mPregapSamples;
}


Q_UINT64 Track::getNumSongSamples() const {
    return mNumSamples;
}


Q_UINT32 Track::getNumPostgapSamples() const {
    return mPostgapSamples;
}


Q_UINT32 Track::getGaplessDataSize() const {
    return mGaplessData;
}


void Track::setSeasonNumber( Q_UINT32 season ) {
    mSeasonNumber = season;
}


Q_UINT32 Track::getSeasonNumber() const {
    return mSeasonNumber;
}


void Track::setEpisodeNumber( Q_UINT32 episode ) {
    mEpisodeNumber = episode;
}


Q_UINT32 Track::getEpisodeNumber() const {
    return mEpisodeNumber;
}


void Track::doneAddingData() {
    // some checks
    if ( getItemProperty( MHOD_ARTIST ).isEmpty() ) {
        setItemProperty( "_no_artist_", MHOD_ARTIST );
    }
    if ( getItemProperty( MHOD_ALBUM ).isEmpty() ) {
        setItemProperty( "_no_album_", MHOD_ALBUM );
    }
    if ( getItemProperty( MHOD_TITLE ).isEmpty() ) {
        setItemProperty( "_no_title_", MHOD_TITLE );
    }

    setArtist( getArtist() );   // refresh artistNameForSort
}


/*** TrackPtrList Implementation details ***/

TrackPtrList::TrackPtrList()
    : itunesdb::utils::SortablePtrVector<itunesdb::Track> ( false )
{
    setDefaultSortOrder();
}


TrackPtrList::~TrackPtrList() {
    // clear(); don't care
}


struct DirtyTrackFinder {
    bool operator () ( const Track * track ) const {
        return track && track->isDirty();
    }
};
bool TrackPtrList::hasDirtyTracks() const {
    return utils::findFirst( begin(), end(), DirtyTrackFinder() ) != end();
}


struct RecentlyPlayedTrackFinder {
    bool operator () ( const itunesdb::Track * track ) const {
        return track->recentlyPlayed();
    }
};
bool TrackPtrList::hasRecentlyPlayedTracks() const {

    return utils::findFirst( begin(), end(),
        RecentlyPlayedTrackFinder() ) != end();

}


TrackPtrList * TrackPtrList::getTracksBy( TrackPredicate& predicate, itunesdb::TrackPtrList& buffer ) const {
    FilteredConstIterator<TrackPredicates::PredicatePtrDelegator> trackIter( *this,
                        TrackPredicates::PredicatePtrDelegator( &predicate ) );
    if ( ! trackIter.hasNext() ) {
        return NULL;
    }
    buffer.addAll( trackIter );
    return &buffer;
}


Track * TrackPtrList::findFirstTrackBy( TrackPredicate& predicate ) const {
    FilteredConstIterator<TrackPredicates::PredicatePtrDelegator> trackIter( *this,
                        TrackPredicates::PredicatePtrDelegator( &predicate ) );
    if ( trackIter.hasNext() ) {
        return trackIter.next();
    } else {
        return NULL;
    }
}


bool TrackPtrList::removeIfTrue( TrackPredicate& predicate ) {
    return filteredElements( TrackPredicates::PredicatePtrDelegator( &predicate ) ).removeRemaining();
}


bool TrackPtrList::removeAllRefs( itunesdb::Track * track ) {
    return filteredElements( std::bind2nd( std::equal_to<itunesdb::Track*>(), track ) ).removeRemaining();
}


class DefaultTrackComparator : public TrackComparator {
public:
    virtual int compare( const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
        int result = 0;
        result = track1.getArtist().localeAwareCompare( track2.getArtist() );
        if ( result ) {
            return result;
        }
        result = track1.getAlbum().localeAwareCompare( track2.getAlbum() );
        if ( result )  {
            return result;
        }
        result = track1.getTrackNumber() - track2.getTrackNumber();
        return result;
    }
};
void TrackPtrList::setDefaultSortOrder() {
    setComparator( new DefaultTrackComparator() );
}


TrackComparatorPtr TrackComparators::createPtr( TrackComparator * comparator ) {
    if ( comparator ) {
        return TrackComparatorPtr( comparator );
    }
    return TrackComparatorPtr( new DefaultTrackComparator() );
}


int TrackComparators::ComposingComparator::compare(
        const itunesdb::Track& track1, const itunesdb::Track& track2 ) const {
    int result = 0;
    ComparatorPtrVector::const_iterator compIter = m_comparators.begin();
    for ( ; compIter != m_comparators.end() && result == 0; ++compIter ) {
        result = (*compIter)->compare( track1, track2 );
    }
    return result;
}


}

